import os
import pkg_resources
from octomy.utils import get_package_file_content

import logging

logger = logging.getLogger(__name__)


version_filename = "VERSION"


def get_version(pkg):
	return get_package_file_content(pkg, version_filename) or "0.0.0"



from pathlib import Path
import pystache


class Svg:
	def __init__(self):
		# Load SVG template from file
		svg_template_fn = "template.svg"
		svg_template_file = Path(svg_template_fn)
		if svg_template_file.is_file():
			with open(svg_template_fn, "r") as svg_template_handle:
				print("Loading svg template from file")
				svg_template_text = svg_template_handle.read()
				# Render the template
				svg_output_fn = "output.svg"
				print("Applying data to template")
				svg_output_text = pystache.render(svg_template_text, {"from": "BGO", "to": "KBP"})
				# Save rendered text into SVG output file
				with open(svg_output_fn, "w") as svg_output_handle:
					print("Writing output svg")
					svg_output_handle.write(svg_output_text)
		else:
			print("Template not found, aboring")



#!/bin/env python3
from functools import lru_cache
from furl import furl
from pydantic import TypeAdapter, ValidationError
from typing import Type, get_args, get_origin
import aiofiles
import asyncio
import datetime
import logging
import octomy.utils
import os
import pathlib
import pprint
import psycopg
import psycopg_pool
import pydantic
import re
import socket
import urllib.parse


logger = logging.getLogger(__name__)


c = octomy.utils.termcol

def schema_to_compact(schema: dict) -> str:
	# Extract the title or name of the model
	model_name = schema.get('title', 'UnknownModel')
	
	# Initialize an empty list to hold field descriptions
	field_descriptions = []
	
	# Iterate over the properties in the schema to build field descriptions
	for field_name, details in schema.get('properties', {}).items():
		field_type = details.get('type', 'UnknownType')
		
		# For more complex types (e.g., arrays, nested models), you might need additional handling here
		if field_type == 'array':
			items = details.get('items')
			if isinstance(items, dict):
				item_type = items.get('type', 'UnknownType')
				field_type = f"List[{item_type}]"
		
		# Append the field description to the list
		field_descriptions.append(f"{field_name}:{field_type}")
	
	# Join all field descriptions into a single string
	fields_str = ', '.join(field_descriptions)
	
	# Return the compact representation of the model
	return f"{model_name}{{ {fields_str} }}"


def describe_type(input_type: Type) -> str:
	# Handle Pydantic models by checking for the .schema() method
	if hasattr(input_type, 'schema') and callable(getattr(input_type, 'schema')):
		schema = schema_to_compact(input_type.schema())
		return f"Pydantic {schema}"
	
	# Check for generic types (e.g., List[SomeType])
	origin_type = get_origin(input_type)
	if origin_type is not None:
		type_args = get_args(input_type)
		type_args_descriptions = [describe_type(t) for t in type_args]
		descs='\n, '.join(type_args_descriptions)
		return f"Generic type: {origin_type.__name__} of [{descs}]"
	# Fallback for standard or simple types
	return f"Type: {input_type}"


def diag_to_str(diag, prefix =""):
	error_attrs = [
		"column_name", "constraint_name", "context", "datatype_name",
		"internal_position", "internal_query", "message_detail", "message_hint",
		"message_primary", "schema_name", "severity", "source_file",
		"source_function", "source_line", "sqlstate", "statement_position",
		"table_name", "severity_nonlocalized"
	]
	out = ""
	for attr in error_attrs:
		value = getattr(diag, attr, None)  # Use getattr to safely handle missing attributes
		if value is not None:  # Only log if the attribute is present and not None
			out += f"{prefix}{attr}: {value}"
	return out

def _q(s):
	if not s:
		return ""
	return urllib.parse.quote(s)

def check_host(hostname, port, timeout_sec = 4):
	try:
		ip_address = socket.gethostbyname(hostname)
		sock = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
		sock.settimeout(timeout_sec)
		result = sock.connect_ex((ip_address, port))
		sock.close()
		if result == 0:
			return True, None
		else:
			return False, f"Port '{port}' is not open on '{hostname}'"
	except socket.gaierror:
		return False, f"Hostname '{hostname}' could not be resolved"
	except socket.timeout:
		return False, f"Connection to the host timed out after {timeout_sec}"
	except Exception as e:
		return False, f"An unexpected error has occurred: {e}"


def verify_db_uri(db_uri, do_online = True):
	def _error(err):
		return None, err
	try:
		result:ParseResult = urllib.parse.urlparse(db_uri)
		#logger.warning(f"{result}")
		if not result:
			return _error(f"DB URI '{db_uri}' could not be pased")
		if result.scheme != 'postgresql' and result.scheme != 'postgres':
			return _error(f"DB URI '{db_uri}' did not have correct schema: '{result.scheme}'")
		if not result.hostname:
			return _error(f"DB URI '{db_uri}' did not have hostname")
		if not result.username:
			return _error(f"DB URI '{db_uri}' did not have username")
		if not result.password:
			return _error(f"DB URI '{db_uri}' did not have password")
		if not result.path.strip('/'):
			return _error(f"DB URI '{db_uri}' did not have database path")
		if do_online:
			try:
				host_ok, host_err = check_host(result.hostname, result.port)
				if not host_ok:
					return None, host_err
				#logger.warning("ONLINE IS GOPING DOWN")
				with psycopg.connect(db_uri) as connection:
					#logger.warning("ONLINE IS GOOD")
					pass
			except psycopg.Error as e:
				diag = diag_to_str(e.diag, prefix = ", ")
				return _error(f"Could not connect with DB URI '{db_uri}': {e} ({type(e)}), diag={diag}, sqlstate={e.sqlstate}, pgconn={e.pgconn}")
	except Exception as e:
		return _error(f"DB URI '{db_uri}' had error while parsing: {e} ({type(e)})")
	return db_uri, None


def db_uri_from_conf(config, do_verify = True, do_online= True, do_redact = False):
	if None == config:
		return None, "No config"
	db_hostname = config.get("db-hostname", None)
	db_port = config.get("db-port", None)
	db_username = config.get("db-username", None)
	db_password = config.get("db-password", None)
	db_database = config.get("db-database", None)
	if do_redact:
		db_password = octomy.utils.redact(db_password)
	f = furl()
	try:
		f.set(scheme="postgres", username = db_username, password = db_password, host = db_hostname, port = db_port, path = db_database)
		if do_verify:
			 return verify_db_uri(f.url, do_online = do_online)
		else:
			return f.url, None
	except Exception as e:
		#logger.error(e, exc_info=e)
		return None, f"Invalid parts: {e} {type(e)}"


def db_uri_to_config(db_uri, do_online = True):
	db_uri_out, err = verify_db_uri(db_uri, do_online = do_online)
	if not db_uri_out:
		#logger.warning(err)
		return {}
	f = furl(db_uri)
	database = (f.path.segments or [""])[0]
	# fmt:off
	config = {
		  "db-hostname": f.host
		, "db-port": f.port
		, "db-username": f.username
		, "db-password": f.password
		, "db-database": database
	}
	# fmt:on
	return config


def get_config_hash(config, prefix = "db-"):
	if not config:
		return ""
	key = ""
	for i in ["hostname", "port", "username", "password", "database"]:
		key += octomy.utils.hashstr(config.get(f"{prefix}{i}", ""))
	return octomy.utils.hashstr(key)


def query_key_to_sql_path(query_key, relative=False, do_debug = True):
	sep = '.'
	if do_debug:
		logger.info(f"                   sep:{sep}")
		logger.info(f"             query_key:{query_key}")
	parts = query_key.split(sep) if sep in query_key else list()
	if do_debug:
		logger.info(f"                 parts: '{pprint.pformat(parts)}'")
	if not parts:
		logger.warning(f"query_key '{query_key}' could not be split")
		return None
	if len(parts) <= 0:
		logger.warning(f"query_key '{query_key}' split into 0")
		return None
	script_name = parts.pop() + ".sql"
	relative_script_path = os.path.join(*parts, "sql", script_name)
	if relative:
		if do_debug:
			logger.info(f"Parsed '{query_key}' to '{relative_script_path}' (relative)")
		return relative_script_path
	package_name = sep.join(parts)
	if do_debug:
		logger.info(f"          package_name:{package_name}")
	package_dir = octomy.utils.get_package_root_dir(package_name)
	if do_debug:
		logger.info(f"  get_package_root_dir:{package_dir}")
	if not package_dir:
		logger.warning(f"get_package_root_dir failed")
		return None
	package_dir = package_dir.parent
	canonical_path = os.path.join(package_dir, relative_script_path)
	canonical_path = pathlib.Path(canonical_path).resolve()
	if do_debug:
		logger.info(f"        canonical_path:{canonical_path}")
	return canonical_path


################################################################################



class DbTime(pydantic.BaseModel):
	now:datetime.datetime


################################################################################


class Database:
	def __init__(self, db_uri):
		self.error = None
		self.db_uri = db_uri
		#self.query_dirs = set()
		self.queries = dict()
		self.pool = None

	def _error(self, message):
		ok = not bool(message)
		# logger.info(f"_error({message}) -> {ok}")
		self.error = None if ok else message
		return ok, self.error

	# Get current time from db
	#def get_now(self, do_debug=False) -> DbTime, str|None:
	#	db_ret, db_err = await self.dbc.query_one("octomy.db.get_now", params=dict(), item_type=DbTime, prepare=False, do_debug=do_debug)
	#	return db_ret, db_err

	async def verify(self, do_online = False):
		try:
			# logger.info(f"VERIFYING '{self.db_uri}'")
			if not self.db_uri:
				return self._error(f"No DB URI: '{self.db_uri}'")
			out, err = verify_db_uri(self.db_uri, do_online = do_online)
			if not out:
				return self._error(err)
			#if not self.query_dirs or len(self.query_dirs) <= 0:
			#	return self._error(f"No query dirs registered: '{self.query_dirs}'")
			return True, None
		except Exception as e:
			logger.warning(f"Unknown error '{e}':", exc_info=True)
			return False, str(e)

	async def state(self, do_online = False, show_full = False):
		head_c=c.GREEN
		logger.info(f"{head_c}Status [{c.ORANGE}{'ONLINE' if do_online else 'OFFLINE'}{head_c}]:")
		ok, err = await self.verify(do_online = do_online)
		if ok:
			logger.info(f"\t{c.GREEN}OK")
		else:
			logger.info(f"\t{c.RED}ERROR: {c.ORANGE}{err}")
		#if self.query_dirs:
		#	logger.info(f"{head_c}Query paths:")
		#	for query_dir in self.query_dirs:
		#		logger.info(f"\t{c.CYAN}{query_dir}")
		#else:
		#	logger.info("No query paths")
		if self.queries:
			logger.info(f"{head_c}Loaded queries:")
			tabs2 = "\t\t"
			for key, query in self.queries.items():
				if show_full:
					logger.info(f"\t{c.BLUE}{key:>10}{c.ENDC}: {c.GREEN}{octomy.utils.indent(query, tabs2)}")
				else:
					logger.info(f"\t{c.BLUE}{key:>10}{c.ENDC}: {c.GREEN}{len(query)} chars")
		else:
			logger.info(f"{head_c}No loaded queries")
		return ok, err

	def reset_queries(self):
		"""Reset the internal cache of queries, forcing queries to be reloaded upon next invocation"""
		self.queries = dict()
		'''
	def preload_queries(self, do_debug = False):
		"""Recursively load and register queries from .sql files in given path."""
		self.reset_queries()
		if do_debug:
			logger.info("Scanning query dirs:")
		for dir in self.query_dirs:
			if not os.path.isdir(dir):
				if do_debug:
					logger.warning(f"Not a dir, skipping: '{dir}'")
					logger.warning(f" * NOTE: CWD is '{os.getcwd()}'")
				continue
			if do_debug:
				logger.info(f" *Scanning query dir: '{dir}'")
			for sql_file in dir.rglob('*.sql'):
				# some/path/file.sql --> some.path.file
				if do_debug:
					logger.info(f" |-Found candidate file: '{sql_file}'")
				query_key = sql_file.relative_to(dir).with_suffix('').as_posix().replace('/', '.')
				with open(sql_file, 'r', encoding='utf-8') as file:
					query = file.read()
					if do_debug:
						#logger.info(f" | |-Loading '{query_key}' from '{sql_file}' with '{query}'")
						logger.info(f" | |-Loading '{query_key}' from '{sql_file}' with query size {len(query)} chars")
					self.queries[query_key] = query

		if do_debug:
			logger.info("Scanning complete")


	def register_query_dir(self, path, do_preload = False, do_debug = False):
		"""Recursively load queries from .sql files in given path."""
		real = os.path.realpath(path)
		if do_debug:
			logger.info(f"Adding query dir: '{path}' ({real})")
		self.query_dirs.add(pathlib.Path(real))
		if do_preload:
			self.preload_queries(do_debug = do_debug)
'''

	async def load_query(self, query_key, do_debug = False):
		"""Load and return a single query by first looking it up in the internal cache and falling back on looking for it on disk"""
		# Cache hit
		if query_key in self.queries:
			return self.queries[query_key]
		# Cache miss, load from disk
		try:
			# Make sure we know where to look
			#if len(self.query_dirs) <= 0:
			#	raise Exception(f"No query dir(s) specified")
			file_path = query_key_to_sql_path(query_key, relative=False, do_debug = do_debug)
			if do_debug:
				logger.info(f"             file_path: '{file_path}'")
			if not file_path:
				raise Exception(f"No path found for {query_key}")
			# Read the SQL query from file
			async with aiofiles.open(file_path, mode='r', encoding='utf-8') as file:
				sql_query = await file.read()
				self.queries[query_key] = sql_query
				return sql_query
			'''
			# Convert the query_key to a file path
			for dir in self.query_dirs:
				file_path = dir / pathlib.Path(query_key.replace('.', '/') + '.sql')
				logger.info(f"file_path {file_path}")
					
				# Read the SQL query from file
				async with aiofiles.open(file_path, mode='r', encoding='utf-8') as file:
					sql_query = await file.read()
				# with open(file_path, 'r', encoding='utf-8') as file:
				#	sql_query = file.read()
				self.queries[query_key] = sql_query
				return sql_query
			'''
		except FileNotFoundError:
			logger.info(f"SQL file {os.path.abspath(file_path)}")
			logger.error(f"SQL file not found for key {query_key} at path {file_path}")
		except Exception as e:
			logger.error(f"{type(e).__name__}: Error loading SQL file for key {query_key}: {e}")
		return None

	def _get_pool(self):
		self.pool = self.pool or psycopg_pool.AsyncConnectionPool(self.db_uri)
		return self.pool

	def _adapt(self, raw, item_type, do_debug=False):
		if not item_type:
			return raw
		if do_debug:
			logger.info(f"Adapting {pprint.pformat(raw)} to {pprint.pformat(item_type)}")
		try:
			return TypeAdapter(item_type).validate_python(raw)
		except Exception as e:
			logger.warning(f"Adaption error details")
			logger.warning(f"                       raw: {pprint.pformat(raw)}")
			logger.warning(f"                 type(raw): {type(raw)}")
			logger.warning(f"                 item_type: {item_type}")
			logger.warning(f"  describe_type(item_type): {describe_type(item_type)}")
			raise

	async def direct_query(self, query, query_name=None, params=dict(), mode="none", item_type=None, prepare=True, do_debug=True):
		cursor = None
		try:
			if not query:
				raise Exception("No query")
			# Make sure we actually have a pool
			pool = self._get_pool()
			if not pool:
				raise Exception("No pool")
			# TODO: Look into implementing pipeline mode
			# See: https://www.psycopg.org/psycopg3/docs/advanced/pipeline.html
			# Acquire a connection from the pool
			async with pool.connection() as connection:
				async with connection.transaction() as transaction:
					async with connection.cursor(row_factory=psycopg.rows.dict_row) as cursor:
						# See https://www.psycopg.org/psycopg3/docs/advanced/prepare.html
						if do_debug:
							logger.info(f"Preparing query '{query_name}'")
							logger.info("\n\n" + query+"\n\n")
							logger.info(f"With data")
							logger.info("\n\n" + pprint.pformat(params)+"\n\n")
						await cursor.execute(query = query, params = params, prepare = prepare)
						if mode == "many":
							res = self._adapt(await cursor.fetchall(), item_type, do_debug)
							if do_debug:
								logger.info(f"Returning many data: '{pprint.pformat(res)}'")
							return res, None
						elif mode == "one":
							res = self._adapt(await cursor.fetchone(), item_type, do_debug)
							if do_debug:
								logger.info(f"Returning one data: '{pprint.pformat(res)}'")
							return res, None
						elif mode == "exists":
							res = await cursor.fetchone()
							if item_type:
								logger.warning(f"item_type speficied '{item_type}' for query with mode exists")
							if do_debug:
								logger.info(f"Returning exist data: '{pprint.pformat(res)}'")
							return res is not None, None
						elif mode == "none":
							if item_type:
								logger.warning(f"item_type speficied '{item_type}' for query with mode none")
							if do_debug:
								logger.info(f"Returning for none")
							return None, None
						else:
							raise Exception(f"Unknown mode {mode}. Should be one of [many, one, exists, none]")
		except Exception as e:
			logger.error("\n\n"+octomy.utils.stripes())
			logger.error("")
			logger.error("")
			logger.error(f"###############################################")
			logger.error(f"+       Error: ({type(e).__name__})")
			logger.error(f"#    Querying: '{query_name}'\n\n{query}\n")
			if cursor and query and "cursor" in query:
				logger.error(f"#    Injected: '{cursor.query.decode()}'")
			logger.error(f"#         On")
			logger.error(f"#         + db_uri =   '{self.db_uri}'")
			logger.error(f"#         For")
			logger.error(f"#         + mode =   '{mode}'")
			logger.error(f"#         + params = '{pprint.pformat(params)}'")
			logger.error(f"# Failed with:")
			if isinstance(e,psycopg.Error):
				logger.error(f"# + e.type:     '{type(e).__name__}'")
				logger.error(f"# + e.sqlstate: '{e.sqlstate}'")
				if getattr(e, "pgconn", None):
					logger.error(f"# + e.pgconn:   '{e.pgconn}'")
				if getattr(e, "pgresult", None):
					logger.error(f"# + e.pgresult: '{e.pgresult}'")
				if e.diag:
					logger.error(f"# + e.diag: '{e.diag}'")
					logger.error(diag_to_str(e.diag, prefix = "#   + "))
			elif isinstance(e, ValidationError):
				logger.error(f"Validation error: with item_type '{item_type}'")
			else:
				# Unexpected errors
				logger.error(f"Unexpected error: {type(e).__name__}")
			logger.error(f"#  Stacktrace: \n\n{e}\n")
			logger.error("\n\n"+octomy.utils.stripes())
			return None, str(e)

	async def key_query(self, query_key, params=dict(), mode="none", item_type=None, prepare=True, do_debug=False):
		query = await self.load_query(query_key)
		if query:
			if do_debug:
				logger.info(f"key_query params='{pprint.pformat(params)}' prepare={prepare}")
			return await self.direct_query(query = query, query_name=query_key, params = params, mode = mode, item_type=item_type, prepare=prepare, do_debug=do_debug)
		return None, "No query"

	async def query_many(self, query_key, params=dict(), item_type=None, prepare=True, do_debug=False):
		if do_debug:
			logger.info(f"query_many params='{pprint.pformat(params)}' prepare={prepare}")
		return await self.key_query(query_key = query_key, params = params, mode = "many", item_type=item_type, prepare=prepare, do_debug=do_debug)

	async def query_one(self, query_key, params=dict(), item_type=None, prepare=True, do_debug=False):
		if do_debug:
			logger.info(f"query_one params='{pprint.pformat(params)}' prepare={prepare}", prepare=True, do_debug=False)
		return await self.key_query(query_key = query_key, params = params, mode = "one", item_type=item_type, prepare=prepare, do_debug=do_debug)

	async def query_none(self, query_key, params=dict(), prepare=True, do_debug=False):
		if do_debug:
			logger.info(f"query_none params='{pprint.pformat(params)}' prepare={prepare}", prepare=True, do_debug=False)
		return await self.key_query(query_key = query_key, params = params, mode = "none", prepare=prepare, do_debug=do_debug)

	async def query_exists(self, query_key, params=dict(), prepare=True, do_debug=False):
		if do_debug:
			logger.info(f"query_exists params='{pprint.pformat(params)}' prepare={prepare}")
		return await self.key_query(query_key = query_key, params = params, mode = "exists", prepare=prepare, do_debug=do_debug)


#@lru_cache
def get_database(config:dict, do_verify=True, do_online = True, do_debug = True):
	uri, err = db_uri_from_conf(config = config, do_verify = do_verify, do_online = do_online, do_redact = False)
	if not uri:
		logger.warning(f"db uri error: {err}")
		return None, err
	db = Database(db_uri = uri)
	if False:
		sql_dir = config.get("db-sql-dir")
		if sql_dir:
			if do_debug:
				logger.info(f"Adding sql dir from config: {sql_dir}")
			db.register_query_dir(sql_dir, do_debug = do_debug)
	return db, None




from enum import Enum
import asyncio
import datetime
import octomy.utils
import logging
import os
import pprint
import re
import subprocess
import sys
import shlex

#1Br_h9krCj2cKFQ8GIo-xBU4ljUvhD7a_
logger = logging.getLogger(__name__)

####################################################### COMMON

# -12345.67890
decimal_number_inc = r'-?\d*\.?\d+'

# "some string"
quoted_string_inc = r'(["\'])(?P<value>(?:(?!\1).)*)\1'

# (mm)
unit_inc = r'(?:\s*\(\s*(?P<unit>[^\s]+)\s*\)\s*)'

####################################################### META

# /* [Section name] */
#h1_re = re.compile(r"^\/\*\s*\[\s*(?P<section>[^\[]+)\s*\]\s*\*\/", flags=re.MULTILINE)
h1_re = re.compile(r"^\/\*\s*\[\s*(?P<section>[^\]\s]+(?:[^\]]*[^\]\s])?)\s*\]\s*\*\/", flags=re.MULTILINE)

# // Variable description
#h2_re = re.compile(r'^\/\/\s*(?P<description>.*)\s*$', flags=re.MULTILINE)
h2_re = re.compile(r'^\s*\/\/\s*(?P<description>[^\s].*?)\s*' + unit_inc + r'?$')

# Variable = definition; // comment
#var_re = re.compile(r'^\s*(?P<variable>\w+)\s*=\s*(?P<value>[^;]+);(?:\s*\/\/\s*(?P<schema>[^\n]*))?\s*$', flags=re.MULTILINE)
var_re = re.compile(r'^\s*(?P<variable>.*?)\s*=\s*(?P<value>.*?)\s*;\s*(?:\s*\/\/\s*(?P<comment>.*?)\s*)?$', flags=re.MULTILINE)


####################################################### VALUES

# true
bool_val_re = re.compile(r'\s*(?P<value>true|false)\s*')

# "mystring"
string_val_re = re.compile(r'\s*' + quoted_string_inc + r'\s*')

# -1234.56789
number_val_re = re.compile(r'\s*(?P<value>' + decimal_number_inc + r')\s*')

# [ 1,2,3,4,5 ]
#vector_val_re = re.compile(r'\s*\[\s*(?P<value>(?:' + decimal_number_inc + r'\s*,?\s*)+)\]\s*')
vector_val_re = re.compile(r'\s*\[\s*(?P<value>' + decimal_number_inc + r'(?:\s*,\s*' + decimal_number_inc + r')*)\]\s*')


####################################################### BOOL SCHEMA


# [ true:label1, false:label2]
#labeled_bool_schema_re = re.compile(r'\s*\[\s*(?P<values>(?:\s*(true|false)\:[\w\s]*)(?:\s*,\s*(true|false):[\w\s]*)*)\s*\]\s*')
labeled_bool_schema_re = re.compile(r'\s*\[\s*(?P<values>(?:\s*(true|false)\s*:\s*\w.*?)(?:\s*,\s*(true|false)\s*:\s*\w.*?)*)\s*\]\s*')


####################################################### STRINGS SCHEMA

# [ value1, value2, value3 ]
values_string_schema_re = re.compile(r'\s*\[\s*(?P<values>(?:[a-zA-Z]\w*)(?:\s*,\s*[a-zA-Z]\w*)*)\s*\]\s*')

# [ value1:label1, value3:label2, value3:label3 ]
#labeled_string_schema_re = re.compile(r'\s*\[\s*(?P<values>(?:[a-zA-Z]\w*\:[\w\s]*)(?:\s*,\s*[a-zA-Z]\w*:[\w\s]*)*)\s*\]\s*')
labeled_string_schema_re = re.compile(r'\s*\[\s*(?P<values>[a-zA-Z]\w*\s*:[^,\s].*?(?:\s*,\s*[a-zA-Z]\w*\s*:[^,\s].*?)*)\s*\]\s*')


####################################################### NUMBERS SCHEMA


# [ max ]
#range1_schema_re = re.compile(r'\s*(?P<max>' + decimal_number_inc + r')\s*\]\s*' + unit_inc + r'?')
range1_schema_re = re.compile(r'\s*\[\s*(?P<max>' + decimal_number_inc + r')\s*\]\s*' + unit_inc + r'?')

# [ min : max ]
range2_schema_re = re.compile(r'\s*\[\s*(?P<min>' + decimal_number_inc + r')\s*:\s*(?P<max>' + decimal_number_inc + r')\s*\]\s*' + unit_inc + r'?')

# [ min : step : max ]
range3_schema_re = re.compile(r'\s*\[\s*(?P<min>' + decimal_number_inc + r')\s*:\s*(?P<step>' + decimal_number_inc + r')\s*:\s*(?P<max>' + decimal_number_inc + r')\s*\]\s*' + unit_inc + r'?')

# [ 123:label1, 456:label2, 789:label3 ]
#labeled_number_schema_re = re.compile(r'\s*\[\s*(?P<values>(?:' + decimal_number_inc + r'\:[\w\s]*)(?:\s*,\s*' + decimal_number_inc + r':[\w\s]*)*)\s*\]\s*' + unit_inc + r'?')
labeled_number_schema_re = re.compile(r'\s*\[\s*(?P<values>(?:' + decimal_number_inc + r'\s*:\s*[^,\s].*?(?:\s*,\s*' + decimal_number_inc + r'\s*:\s*[^,\s].*?)*))\s*\]\s*' + unit_inc + r'?')

# [ 123, 456, 789 ]
values_number_schema_re = re.compile(r'\s*\[\s*(?P<values>' + decimal_number_inc + r'(?:\s*,\s*' + decimal_number_inc + r')*)\s*\]\s*' + unit_inc + r'?')


#######################################################

version_re = re.compile(r'OpenSCAD version\s*(?P<version>.*)\s*')

def parse_version(text):
	m = version_re.match(text)
	if m:
		g = m.groupdict()
		return g.get("version")
	return None


#######################################################



def now():
	return datetime.datetime.now(datetime.timezone.utc)

def millinow():
	tz = datetime.timezone.utc
	now = datetime.datetime.now(tz)
	epoch = datetime.datetime(1970, 1, 1, tzinfo = tz) # use POSIX epoch
	posix_timestamp_millis = (now - epoch) // datetime.timedelta(milliseconds=1) 
	return posix_timestamp_millis

def shex(raw):
	neat = shlex.quote(raw)
	if raw != neat:
		logger.info(f"SHEX {raw} -> {neat}")
	return neat

def parse_number(num_str):
	if not num_str:
		 # logger.warning(f"No string '{num_str}' to convert to number, returning 0");
		return 0
	num_str = num_str.strip().lower()
	try:
		num_int = int(num_str)
		#if str(num_int) == num_str:
		return num_int
	except:
		pass
	try:
		num_float = float(num_str)
		#if str(num_float) == num_str:
		return num_float
	except:
		pass
	logger.warning(f"Could not convert string '{num_str}' to number, returning 0");
	return 0

def parse_vector(raw):
	# logger.info(" parse_vector()")
	out = list()
	if raw:
		# logger.info(f"VECTOR RAW: {pprint.pformat(raw)}")
		for part in raw.split(","):
			# logger.info(f"VECTOR PART: {pprint.pformat(part)}")
			out.append(parse_number(part))
	return out


def parse_value(raw, do_debug = False):
	if do_debug:
		logger.info(f"parse_value({raw})")
	if raw:
		text = raw.strip()
		m = bool_val_re.match(text)
		if m:
			ret={ "type":"bool", "value": "true" == m.groupdict().get("value", "false").strip().lower()}
			if do_debug:
				logger.info(f"MATCHED BOOL: {pprint.pformat(ret)}")
			return  ret
		m = string_val_re.match(text)
		if m:
			ret = { "type":"string", "value": m.groupdict().get("value", "")}
			if do_debug:
				logger.info(f"MATCHED STRING: {pprint.pformat(ret)}")
			return  ret
		m = number_val_re.match(text)
		if m:
			ret = { "type":"number", "value": parse_number(m.groupdict().get("value", 0))}
			if do_debug:
				logger.info(f"MATCHED NUMBER: {pprint.pformat(ret)}")
			return  ret
		m = vector_val_re.match(text)
		if m:
			# logger.info("------------------")
			# pprint.pformat(m.groupdict())
			# logger.info("------------------")
			v = m.groupdict().get("value")
			# logger.info(f"BALLLLLLS {v}")
			ret = { "type":"vector", "value": parse_vector(v)}
			if do_debug:
				logger.info(f"MATCHED VECTOR: {pprint.pformat(ret)}")
			return  ret
	else:
		logger.warn(f"NO RAW '{raw}'")
	if do_debug:
		logger.info(f"NO MATCH FOR {raw}")
	return dict()


def parse_range_schema(text, do_debug = False):
	if do_debug:
		logger.info(" parse_range_schema()")
	ret = {"raw":text}
	m = range1_schema_re.match(text) or range2_schema_re.match(text) or range3_schema_re.match(text)
	if m:
		g = m.groupdict()
		if do_debug:
			logger.info(f"   + range {g}")
		ret["min"] = parse_number(g.get("min"))
		ret["step"] = parse_number(g.get("step"))
		ret["max"] = parse_number(g.get("max"))
		ret["unit"] = str(g.get("unit"))
		if do_debug:
			logger.info(f"   + range {ret['min']}, {ret['step']}, {ret['max']}:")
	return ret


def parse_number_values_schema(text, do_debug = False):
	# logger.info(" parse_number_values_schema()")
	ret = {"raw":text}
	# Labels
	m = labeled_number_schema_re.match(text)
	if m:
		g = m.groupdict()
		#logger.info(f"   + values {g}")
		vs = g.get("values")
		if vs:
			labels = dict()
			for pair in [ v.strip().split(":") for v in vs.split(",")]:
				labels[parse_number(pair[0])] = pair[1]
			ret["labels"] = labels
		else:
			#logger.info(f"   + vs {pprint.pformat(vs)}")
			pass
		# logger.info(f"   + string {pprint.pformat(ret)}")
	else:
		# Values
		m = values_number_schema_re.match(text)
		if m:
			g = m.groupdict()
			#logger.info(f"   + values {g}")
			vs = g.get("values")
			if vs:
				ret["values"] = [ parse_number(v) for v in vs.split(",")]
			else:
				#logger.info(f"   + vs {pprint.pformat(vs)}")
				pass
			# logger.info(f"   + string {pprint.pformat(ret)}")
		else:
			logger.warning(f"Problem parsing values for: {text}")
	return ret

def parse_number_schema(text):
	# logger.info(" parse_number_schema()")
	ret = parse_range_schema(text)
	if not ret.get("max"):
		# logger.warning(f"No range found, trying values for '{text}'")
		ret = parse_number_values_schema(text)
	return ret

def parse_vector_schema(text):
	# logger.info(" parse_vector_schema()")
	return parse_range_schema(text)

def parse_string_schema(text):
	# logger.info(" parse_string_schema()")
	ret = {"raw":text}
	# Labels
	m = labeled_string_schema_re.match(text)
	if m:
		g = m.groupdict()
		#logger.info(f"   + values {g}")
		vs = g.get("values")
		if vs:
			labels = dict()
			for pair in [ v.strip().split(":") for v in vs.split(",")]:
				labels[pair[0]] = pair[1]
			ret["labels"] = labels
		else:
			#logger.info(f"   + vs {pprint.pformat(vs)}")
			pass
		# logger.info(f"   + string {pprint.pformat(ret)}")
	else:
		# Values
		m = values_string_schema_re.match(text)
		if m:
			g = m.groupdict()
			#logger.info(f"   + values {g}")
			vs = g.get("values")
			if vs:
				ret["values"] = [ v.strip() for v in vs.split(",")]
			else:
				#logger.info(f"   + vs {pprint.pformat(vs)}")
				pass
			# logger.info(f"   + string {pprint.pformat(ret)}")
	return ret

def parse_bool_schema(text):
	# logger.info(" parse_bool_schema()")
	ret = {"raw":text}
	# Labels
	m = labeled_bool_schema_re.match(text)
	if m:
		g = m.groupdict()
		#logger.info(f"   + values {g}")
		vs = g.get("values")
		if vs:
			labels = dict()
			for pair in [ v.strip().split(":") for v in vs.split(",")]:
				labels[pair[0]] = pair[1]
			ret["labels"] = labels
		else:
			#logger.info(f"   + vs {pprint.pformat(vs)}")
			pass
		# logger.info(f"   + string {pprint.pformat(ret)}")
	return ret


def parse_schema(raw, type):
	# logger.info(" parse_schema()")
	if raw:
		#logger.info(f"RAW {raw}:")
		text = raw.strip()
		if type == "number":
			return parse_number_schema(text)
		elif type == "vector":
			return parse_vector_schema(text)
		elif type == "string":
			return parse_string_schema(text)
		elif type == "bool":
			return parse_bool_schema(text)
		return {"debug":f"NO SCHEMA FOR {raw} with type '{type}'"}
	else:
		return {"debug":f"NO RAW {raw} with type '{type}'"}

def parse_customize(source, do_debug = False):
	#logger.info(" parse_customize()")
	source = str(source)
	lines = source.split("\n")
	last_h1 = None
	last_h2 = None
	last_unit = None
	parameters_list = list()
	start = now()
	for line in lines:
		line=line.strip()
		if line =="":
			continue
		if do_debug:
			logger.info("======================================")
		t = "?"
		if line.startswith("$"):
			t="$"
		m = var_re.match(line)
		if m:
			t = "VAR"
			g = m.groupdict()
			variable = g.get('variable')
			value = g.get('value')
			default = parse_value(value, do_debug = do_debug)
			type = default.get("type")
			comment = g.get('comment')
			schema = parse_schema(comment, type)
			# Clean up
			variable = variable and variable.strip()
			value = value and value.strip()
			default_value = default.get("value")
			# Make sure that values contain the default value for consistency
			values = schema and schema.get("values")
			if values:
				if None != default_value and not default_value in values:
					values.append(default_value)
					schema["values"] = values
			# Make sure that labels contain the default value for consistency
			labels = schema and schema.get("labels")
			if labels:
				if None != default_value and not default_value in labels:
					labels[default_value] = "default"
					schema["labels"] = labels
				label = labels.get(default_value)
				#schema["label"] = label
			var = {
				  "name": variable
				, "description": last_h2 and last_h2.strip()
				, "section": last_h1 and last_h1.strip()
				, "default": default
				, "unit": last_unit
				, "schema": schema
			}
			parameters_list.append(var)
			# Avoid re-using unit and section
			last_unit = None
			last_h2 = None
		else:
			m = h2_re.match(line)
			if m:
				g = m.groupdict()
				t=" H2"
				last_h2 = g.get("description")
				last_unit = g.get("unit")
			else:
				m = h1_re.match(line)
				if m:
					t=" H1"
					g = m.groupdict()
					name = g.get("section")
					last_h1 = name and name.strip()
				else:
					pass
					#break;
		if do_debug:
			logger.info(f"{t}: {line}")
		if "?" == t:
			# End of initial variable section
			break;
	took = now() - start
	if do_debug:
		logger.info("#"*240)
		logger.info("\n"*2)
	#logger.info(f"Parsing took {octomy.utils.human_delta(took)}")
	#logger.info(f"Final parsed parameters for source {octomy.utils.human_bytesize(len(source))} ({len(lines)} lines):")
	#logger.info(f"parameters:{len(parameters_list)}")
	#logger.info(f"D:{pprint.pformat(parameters_list)}")
	#logger.info("...")
	return parameters_list


class RenderMode(Enum):
	"""
	RenderMode Enum class.
	- test_only
	- render
	- preview
	- thrown_together
	- wireframe
	"""
	test_only = "Test"
	render = "Render"
	preview = "Preview"
	thrown_together = "Thrown Together"
	wireframe = "Wireframe"


class ColorScheme(Enum):
	"""
	ColorScheme Enum class.
	- cornfield
	- metallic
	- sunset
	- starnight
	- beforedawn
	- nature
	- deepocean
	- solarized
	- tomorrow
	- tomorrow_night
	- monotone
	"""
	cornfield      = "Cornfield"
	metallic       = "Metallic"
	sunset         = "Sunset"
	starnight      = "Starnight"
	beforedawn     = "BeforeDawn"
	nature         = "Nature"
	deepocean      = "DeepOcean"
	solarized      = "Solarized"
	tomorrow       = "Tomorrow"
	tomorrow_night = "Tomorrow Night"
	monotone       = "Monotone"

# From https://stackoverflow.com/questions/377017/test-if-executable-exists-in-python
def which(program):
	def is_exe(fpath):
		return os.path.isfile(fpath) and os.access(fpath, os.X_OK)

	fpath, fname = os.path.split(program)
	if fpath:
		if is_exe(program):
			return program
	else:
		for path in os.environ.get("PATH", "").split(os.pathsep):
			exe_file = os.path.join(path, program)
			if is_exe(exe_file):
				return exe_file

	return None

class OpenScadRunner(object):
	def __init__(
		  self
		, scriptfile
		, outfiles
		, version = False
		, animate = None
		, animate_duration = 250
		, antialias = 1.0
		, auto_center = False
		, camera = None
		, color_scheme = ColorScheme.cornfield
		, csg_limit = None
		, customizer_file = None
		, customizer_params = dict()
		, customizer_sets = list()
		, deps_file = None
		, hard_warnings = False
		, imgsize = (1024, 1024)
		, make_file = None
		, orthographic = False
		, quiet = False
		, render_mode = RenderMode.render
		, set_vars = dict()
		, show_axes = True
		, show_crosshairs = False
		, show_edges = False
		, show_scales = True
		, verbose = False
		, view_all = False
		):
		
		"""
		Initializer method.  Arguments are:
		- scriptfile = The name of the script file to process.
		- outfiles = A list of output files to produce. The filetype  for each file is deduced by the extensions.
		- version = If enabled, the program will return version as a string and quit, ignoring all other parameters
		- animate = If given an integer number of frames, creates that many frames of animation, and collates them into an animated GIF.  Default: None
		- animate_duration = Number of milliseconds per frame for an animated GIF.  Default: 250
		- antialias = The antialiasing scaling factor.  If greater than 1.0, images are generated at a larger size, then scaled down to the target size with anti-aliasing.  Default: 1.0  (no anti-aliasing)
		- auto_center = If True, and script does not set $vpt, $vpr, or $vpd, then centers the shape in the rendered image.  Default: False
		- camera = Gives the camera position as either [translate_x,y,z,rot_x,y,z,dist] or [eye_x,y,z,center_x,y,z]
		- color_scheme = The color scheme to render an image with.  See ColorScheme Enum.  Default: ColorScheme.cornfield,
		- csg_limit = If given, the maximum number of CSG elements to render.
		- customizer_file = If given, specifies the file containing Customizer Parameters.
		- customizer_params = An optional list of parameter to set directly
		- customizer_sets = A list of the parameter sets taht should be used from customizer_file.
		- deps_file = If given, the file to write Makefile dependancies out to.
		- hard_warnings = Stop at first WARNING, as if it were an ERROR.  Default: False
		- imgsize = The size of the imagefile to output to, if outputting to a PNG or GIF.  Default: (640,480)
		- make_file = If given, the Makefile script to run when missing a dependency.
		- orthographic = If True, render orthographic.  If False, render with perspective.  Default: False
		- predefined_variables = An optional set of variables to set
		- quiet = Suppresses non-error, non-warning messages.  Default: False
		- render_mode = The rendering mode to use when generating an image.  See RenderMode Enum.  Default: RenderMode.render
		- set_vars = An optional dictionary of script variables and values to set.
		- show_axes = If True, show axes in the rendering.  Default: True
		- show_crosshairs = If True, shows the crosshairs for the center of the camera translation.  Default: False
		- show_edges = If True, shows the edges of all the faces.  Default: False
		- show_scales = If True, show the scales along the axes.  Default: True
		- verbose = Print the command-line to stdout on each execution.  Default: False
		- view_all = If True, and script does not set $vpd, then the field of view is scaled to show the complete rendered shape.  Default: False
		"""
		self.command = "/usr/bin/openscad"
		self.scriptfile = scriptfile
		self.outfiles = outfiles
		self.version = version
		self.animate = animate
		self.animate_duration = animate_duration
		self.antialias = antialias
		self.auto_center = auto_center
		self.camera = camera
		self.color_scheme = color_scheme
		self.csg_limit = csg_limit
		self.customizer_file = customizer_file
		self.customizer_params = customizer_params
		self.customizer_sets = customizer_sets
		self.deps_file = deps_file
		self.hard_warnings = hard_warnings
		self.imgsize = imgsize
		self.make_file = make_file
		self.orthographic = orthographic
		self.quiet = quiet
		self.render_mode = render_mode
		self.set_vars = set_vars
		self.show_axes = show_axes
		self.show_crosshairs = show_crosshairs
		self.show_edges = show_edges
		self.show_scales = show_scales
		self.summary = "all"
		self.verbose = verbose
		self.view_all = view_all
		self.timeout = 2*60
		self.reset()

	def reset(self):
		self.cmdline = []
		self.complete = False
		self.echos = []
		self.errors = []
		self.return_code = None
		self.script = []
		self.stderr = []
		self.stdout = []
		self.success = False
		self.warnings = []

	def __bool__(self):
		"""
		Returns True if the run() method has been called, and the processing is complete, whether or not it was successful.
		"""
		return self.complete

	def good(self):
		"""
		Returns True if the run() method has been called, and the result was successful.
		"""
		return self.success

	def _process_frame(self, imgfile):
		img = Image.open(imgfile)
		img.thumbnail(self.imgsize, Image.ANTIALIAS)
		os.unlink(imgfile)
		img.save(imgfile)
		
	def _process_animation(self, basename):
		imgfiles = ["{}{:05d}.png".format(basename,i) for i in range(self.animate)]
		if fileext == ".gif":
			imgs = []
			for imgfile in imgfiles:
				img = Image.open(imgfile)
				if self.antialias != 1.0:
					img.thumbnail(self.imgsize, Image.ANTIALIAS)
				imgs.append(img)
			imgs[0].save(
				self.outfile,
				save_all=True,
				append_images=imgs[1:],
				duration=self.animate_duration,
				loop=0
			)
			pygifsicle.optimize(self.outfile, colors=64)
		elif fileext == ".png":
			if self.antialias != 1.0:
				for imgfile in imgfiles:
					self._process_frame(imgfile)
			APNG.from_files(imgfiles, delay=self.animate_duration).save(self.outfile)
		for imgfile in imgfiles:
			os.unlink(imgfile)

	def _check_command_availability(self):
		w = which(self.command)
		return w and True or False

	def _prepare_cmd(self):
		self.foo_file = "foo.term"
		outfiles = self.outfiles
		"""
		basename, fileext = os.path.splitext(outfile)
		fileext = fileext.lower()
		if self.animate is not None:
			assert (fileext in (".gif", ".png")), "Can only animate to a gif or png file."
			basename = basename.replace(".", "_")
			outfile = basename + ".png"
		"""
		if self.version:
			self.command_line = [self.command, "--version"]
			return
		if self.animate is not None:
			logger.error("Animation not supported at this time, aborting...")
			return None

		if self.render_mode == RenderMode.test_only:
			self.command_line = [self.command, "-o", self.foo_file]
		else:
			self.command_line = [self.command]
			if isinstance(outfiles, str):
				outfiles = [ outfiles]
			if type(outfiles) is not list:
				logger.error("Malformed outfiles: {pprint.pformat(outfiles)}")
				return None
			for outfile in outfiles:
				self.command_line.extend(["-o", shex(outfile)])
			"""
			if fileext in [".png", ".gif"]:
				self.command_line.append(f"--imgsize={int(self.imgsize[0] * self.antialias)},{int(self.imgsize[1] * self.antialias)}")
			"""
			if self.imgsize:
				#self.command_line.append(f"--imgsize={int(self.imgsize[0] * self.antialias)},{int(self.imgsize[1] * self.antialias)}")
				self.command_line.extend(["--imgsize", f"{int(self.imgsize[0] * self.antialias)},{int(self.imgsize[1] * self.antialias)}"])
			if self.show_axes or self.show_scales or self.show_edges or self.show_crosshairs or self.render_mode==RenderMode.wireframe:
				showparts = []
				if self.show_axes:
					showparts.append("axes")
				if self.show_scales:
					showparts.append("scales")
				if self.show_edges:
					showparts.append("edges")
				if self.show_crosshairs:
					showparts.append("crosshairs")
				if self.render_mode == RenderMode.wireframe:
					showparts.append("wireframe")
				self.command_line.append("--view=" + ",".join(showparts))
			if self.camera is not None:
				while len(self.camera) < 6:
					self.camera.append(0)
				self.command_line.extend(["--camera", ",".join(str(x) for x in self.camera)])
			if self.color_scheme != ColorScheme.cornfield:
				self.command_line.extend(["--colorscheme", self.color_scheme])
			self.command_line.append("--projection=o" if self.orthographic else "--projection=p")
			if self.auto_center:
				self.command_line.append("--autocenter")
			if self.view_all:
				self.command_line.append("--viewall")
			if self.animate is not None:
				self.command_line.extend(["--animate", str(self.animate)])
			if self.render_mode == RenderMode.render:
				self.command_line.extend(["--render", ""])
			elif self.render_mode == RenderMode.preview:
				self.command_line.extend(["--preview", ""])
			elif self.render_mode == RenderMode.thrown_together:
				self.command_line.extend(["--preview", "throwntogether"])
			elif self.render_mode == RenderMode.wireframe:
				self.command_line.extend(["--render", ""])
			if self.csg_limit is not None:
				self.command_line.extend(["--csglimit", self.csg_limit])
		if self.deps_file != None:
			self.command_line.extend(["-d", self.deps_file])
		if self.summary:
			pass #self.command_line.extend(["--summary", self.summary])
		# https://ochafik.com/jekyll/update/2022/02/09/openscad-fast-csg-contibution.html
		#self.command_line.extend(["--enable", "lazy-union"])
		#self.command_line.extend(["--enable", "fast-csg"])
		if self.make_file != None:
			self.command_line.extend(["-m", self.make_file])
		for var, val in self.set_vars.items():
			self.command_line.extend(["-D", f"{var}={val}"])
		if self.customizer_file is not None:
			self.command_line.extend(["-p", shex(self.customizer_file)])
		for var, val in self.customizer_params.items():
			self.command_line.extend(["-D", f"{shex(var)}={shex(val)}"])
		for setName in self.customizer_sets:
			self.command_line.extend(["-P", f"{shex(setName)}"])
		if self.hard_warnings:
			self.command_line.append("--hardwarnings")
		if self.quiet:
			self.command_line.append("--quiet")
		self.command_line.append(shex(self.scriptfile))
		if self.verbose:
			line = " ".join([
				f"'{arg}'" if ' ' in arg or arg == '' else arg
				for arg in self.command_line
			])
			logger.info("Commandline:")
			logger.info(line)

	def to_string(self):
		pass

	def error_string(self):
		return "\nERR:".join(self.errors) + "\nWARN:".join(self.warnings)+ "\nSTDERR:".join(self.stderr)+ "\nCMD:".join(self.command_line)


	async def _async_run(self):
		# https://docs.python.org/3/library/asyncio-subprocess.html#examples
		# coroutine asyncio.create_subprocess_exec(
		#	program
		#	, *args
		#	, stdin=None
		#	, stdout=None
		#	, stderr=None
		#	, limit=None
		#	, **kwds)

		args = list(self.command_line)
		program = args.pop()
		logger.info("Program: {program}")
		logger.info("Args: {len(args)}")
		proc = await asyncio.create_subprocess_exec(
			program = program
			, args = args
			, stdin = asyncio.subprocess.PIPE
			, stdout = asyncio.subprocess.PIPE
			, stderr = asyncio.subprocess.PIPE
			, close_fds = True)
	
		# Read one line of output.
		data = await proc.stdout.readline()
		line = data.decode('ascii').rstrip()
	
		# Wait for the subprocess exit.
		await proc.wait()
		return line
	
		date = asyncio.run(get_date())
		print(f"Current date: {date}")
		

	def _sync_run(self):
		try:
			proc = subprocess.Popen(self.command_line, shell=False, stdin=subprocess.PIPE, stdout=subprocess.PIPE, stderr=subprocess.PIPE, close_fds=True)
			(stdoutdata, stderrdata) = proc.communicate(timeout=self.timeout)
			return stdoutdata, stderrdata, proc.returncode
		except subprocess.TimeoutExpired as te:
			logger.error(f"Command timed out after {self.timeout} seconds")
			return b"", b"", -1

	def run(self):
		"""
		Runs the OpenSCAD application with the current paramaters.
		"""
		try:
			start = now()
			self.reset()
			if not self._check_command_availability():
				self.errors = [f"Command not found or not executable ('{self.command}') "]
				self.success = False
				return self.success, self.error_string()
			self._prepare_cmd()
			stdoutdata = ""
			stderrdata = ""
			returncode = None
			self.do_async = False
			if self.do_async:
				stdoutdata, stderrdata, returncode = self._async_run()
			else:
				stdoutdata, stderrdata, returncode = self._sync_run()
			stdoutdata = stdoutdata.decode('utf-8')
			stderrdata = stderrdata.decode('utf-8')
			if self.verbose:
				logger.info("RAW STDOUT:")
				logger.info(stdoutdata)
				logger.info("RAW STDERR:")
				logger.info(stderrdata)
			self.return_code = returncode
			self.cmdline = self.command_line
			self.stderr = stderrdata.split("\n")
			self.stdout = stdoutdata.split("\n")
			self.echos    = [x for x in self.stderr if x.startswith("ECHO:")]
			self.warnings = [x for x in self.stderr if x.startswith("WARNING:")]
			self.errors   = [x for x in self.stderr if x.startswith("ERROR:") or x.startswith("TRACE:")]
			if self.version:
				self.version = parse_version(stderrdata)
				logger.info(f"PARSED VERSION: {stderrdata} into {self.version}")
			# OLD SUCCESS:
			"""
			if self.return_code == 0 and self.errors == [] and (not self.hard_warnings or self.warnings == []):
				self.success = True
			"""
			if self.return_code == 0:
				self.success = True
			if self.render_mode == RenderMode.test_only and os.path.isfile(self.foo_file):
				os.unlink(self.foo_file)
			#with open(self.scriptfile, "r") as f:
			#	self.script = f.readlines();
			if self.success and self.render_mode != RenderMode.test_only:
				if not self.animate:
					if float(self.antialias) != 1.0:
						for outfile in self.outfiles:
							self._process_frame(outfile)
				else:
					self.process_animation(basename)
			self.complete = True
			took = now() - start
			logger.info(f"Rendering took {octomy.utils.human_delta(took)}")
			return self.success, self.success and None or self.error_string()
		except Exception as e:
			logger.exception("Error occurred while running OpenScad")
			return self.success, f"Exception: '{pprint.pformat(e)}'"
		# Whatever happened here, tell the system
		return self.success, "Unexpected end of run()"



import datetime
import logging
import octomy.config
import octomy.db
import pydantic

logger = logging.getLogger(__name__)



class DbTime(pydantic.BaseModel):
	now:datetime.datetime

class DatabaseCheck:
	def __init__(self, config):
		config:octomy.config.OctomyConfig = octomy.config.get_config()
		self.dbc:octomy.db.Database = octomy.db.get_database(config)
		assert self.dbc.is_ok()
		self.create_tables()


	# Get current time from db
	def get_now(self, do_debug=False) -> DbTime, str|None:
		db_ret, db_err = await self.dbc.query_One("octomy.db.get_now", params=dict(), item_type=DbTime, prepare=False, do_debug=do_debug)
		return db_ret, db_err
 

	def verify(self):
		try:
			db_ret, db_err = self.get_now()
			return None == db_err
		except Exception as e:
			logger.warning(f"Unknown error '{e}':", exc_info=True)
			return False



from octomy import utils
from octomy.cad import openscad
from octomy.cad.generators import file_properties
from octomy.cad.types import *
import datetime
import jinja2
import json
import logging
import os
import pprint
import random
import re

logger = logging.getLogger(__name__)


def to_openscad_parameters(parameters:GeneratorParameters, set_name = "set", do_debug = False):
	out = dict()
	for p in parameters.parameters:
		out[p.name] = p.value
	out = {
		"fileFormatVersion": "1"
		, "parameterSets": {
			set_name: out
		}
	}
	if do_debug:
		logger.info("parameters in openscad format:")
		logger.info(pprint.pformat(out))
	return out


def millinow():
	tz = datetime.timezone.utc
	now = datetime.datetime.now(tz)
	epoch = datetime.datetime(1970, 1, 1, tzinfo = tz) # use POSIX epoch
	posix_timestamp_millis = (now - epoch) // datetime.timedelta(milliseconds = 1) 
	return posix_timestamp_millis


def generate_local_temp_names(id):
	# Add millisecond timestamp to avoid problems in the (rare) case where two calls are made for the same file at the same time
	now = millinow()
	local_name_base = f"{id}_{now}"
	local_name_model = f"/tmp/{local_name_base}.openscad"
	local_name_stl = f"/tmp/{local_name_base}.stl"
	local_name_json = f"/tmp/{local_name_base}.json"
	local_name_log = f"/tmp/{local_name_base}.log"
	local_name_png = f"/tmp/{local_name_base}.png"
	return now, local_name_base, local_name_model, local_name_stl, local_name_json, local_name_log, local_name_png

def generate_derivative_filenames(id, parameters):
	if not parameters:
		logger.warning(f"No parameters ({parameters}) provided while generating derivative filenames for '{id}'")
		return None, None, None, None, None, None
	key = utils.calculate_unique_key(parameters.json())
	drive_name_base = f"{id}_{key}"
	drive_name_stl = f"{drive_name_base}.stl"
	drive_name_json = f"{drive_name_base}.json"
	drive_name_png = f"{drive_name_base}.png"
	drive_name_log = f"{drive_name_base}.log"
	return key, drive_name_base, drive_name_stl, drive_name_json, drive_name_png, drive_name_log


def _prepare_log(cmdline, stdout, stderr, returncode):
	cmdline = " \\\n\t".join(cmdline)
	stdout = "\n".join(stdout)
	stderr = "\n".join(stderr)
	return f"""
---
title: commandline
returncode: {returncode}
---		
{cmdline}

---
title: stdout
---
{stdout}

---
title: stderr
---
{stderr}


	"""

class OpenSCADGenerator:
	def __init__(self, common):
		self.common = common
		self.version = None
		
	def _get_version_worker(self):
		openscad_parameters = {
			"version": True
		}
		oscad = openscad.OpenScadRunner(
			scriptfile = None
			, outfiles = None
			, **openscad_parameters)
		ok, err = oscad.run()
		if ok:
			return oscad.version
		else:
			logger.warning(f"Could not get version: {err}")
		return None

	def get_version(self) -> str:
		if not self.version:
			self.version = self._get_version_worker()
		return self.version and str(self.version) or None
		

	def get_schema(self, source, do_debug = False):
		#logger.info(f"PARSING SCHEMA FROM SOURCE: '{source}':")
		schema = openscad.parse_customize(source = source, do_debug = do_debug)
		#logger.info(pprint.pformat(schema))
		return schema


	def generate(self, source_name, source, parameters, id, do_stl=True, do_png=True, do_debug = False):
		set_name = "set"
		image_sz = 1600;
		openscad_parameters = to_openscad_parameters(parameters, set_name)
		json_parameters = json.dumps(openscad_parameters, indent=3)
		now, local_name_base, local_name_model, local_name_stl, local_name_json, local_name_log, local_name_png = generate_local_temp_names(id)
		key, drive_name_base, drive_name_stl, drive_name_json, drive_name_png, drive_name_log = generate_derivative_filenames(id, parameters)
		with open(local_name_json, "w") as f:
			f.write(json_parameters)
		openscad_parameters = {
			  "animate": None
			, "animate_duration": 250
			, "antialias": 1.0
			, "auto_center": True
			, "camera": None
			, "color_scheme": openscad.ColorScheme.cornfield
			, "csg_limit": None
			, "customizer_file": local_name_json
			, "customizer_params": dict()
			, "customizer_sets": [set_name]
			, "deps_file": None
			, "hard_warnings": False
			, "imgsize": (1600,1600)
			, "make_file": None
			, "orthographic": False
			, "quiet": False
			, "render_mode": openscad.RenderMode.render
			, "set_vars": dict()
			, "show_axes": True
			, "show_crosshairs": False
			, "show_edges": False
			, "show_scales": True
			, "verbose": do_debug or True
			, "view_all": True
		}
		outfiles = list()
		stl_file_id = None
		json_file_id = None
		log_file_id = None
		png_file_id = None

		with open(local_name_model, "wb") as f:
			f.write(source)
		source_folder_id = self.common.ensure_source_folder(source_name)
		part_folder_id = self.common.ensure_part_folder(source_folder_id, key)
	
		if do_stl:
			outfiles.append(local_name_stl)
		if do_png:
			outfiles.append(local_name_png)
		if len(outfiles) > 0:
			oscad = openscad.OpenScadRunner(
				scriptfile = local_name_model
				, outfiles = outfiles
				, **openscad_parameters)
			ok, err = oscad.run()
			if ok:
				props = file_properties(id, key)
				log = _prepare_log(oscad.cmdline, oscad.stdout, oscad.stderr, oscad.return_code)
				with open(local_name_log, "w") as f:
					f.write(log)
				if do_stl:
					stl_file_id = self.common.upload_derivative(local_name_stl, part_folder_id, drive_name_stl, MIMETYPE_STL_MODEL, props)
					json_file_id = self.common.upload_derivative(local_name_json, part_folder_id, drive_name_json, MIMETYPE_JSON_DATA, props)
					log_file_id = self.common.upload_derivative(local_name_log, part_folder_id, drive_name_log, MIMETYPE_TEXT, props)
				if do_png:
					png_file_id = self.common.upload_derivative(local_name_png, part_folder_id, drive_name_png, MIMETYPE_PNG_IMAGE, props)
					if stl_file_id:
						self.common.set_drive_thumbnail_from_local_path(stl_file_id, local_name_png)
			else:
				logger.warning("The parameters were:")
				logger.warning(json_parameters)
				logger.error(f"ERROR(S) OCCURRED for oscad run: {err}")
				utils.delete_file(local_name_json)
				utils.delete_file(local_name_stl)
				utils.delete_file(local_name_model)
				return None, {"message":"Could not generate stl from source with OpenSCAD", "error": oscad.error_string()}
		# Clean up local temp files
		utils.delete_file(local_name_json)
		utils.delete_file(local_name_model)
		utils.delete_file(local_name_png)
		utils.delete_file(local_name_stl)
		if do_debug:
			logger.info(f"Returning: {stl_file_id}")
		return {
			  "model_id": stl_file_id
			, "parameters_id": json_file_id
			, "log_id": log_file_id
			, "thumbnail_id": png_file_id
		}, None



from __future__ import print_function
from queue import Queue
from threading import Thread, Event, Lock
from sys import stdout, stderr
from time import sleep
import sys


def eprint(*args, **kwargs):
	print(*args, file=sys.stderr, **kwargs)


# default exception handler. if you want to take some action on failed tasks
# maybe add the task back into the queue, then make your own handler and pass it in
def default_handler(name, exception, *args, **kwargs):
	eprint(f"{name} raised {exception} with args {repr(args)} and kwargs {repr(kwargs)}")
	pass


# class for workers
class Worker(Thread):

	"""Thread executing tasks from a given tasks queue"""

	def __init__(self, name, queue, results, abort, idle, exception_handler):
		Thread.__init__(self)
		self.name = name
		self.queue = queue
		self.results = results
		self.abort = abort
		self.idle = idle
		self.exception_handler = exception_handler
		self.daemon = True
		self.start()

	"""Thread work loop calling the function with the params"""

	def run(self):
		# keep running until told to abort
		while not self.abort.is_set():
			try:
				# get a task and raise immediately if none available
				func, args, kwargs = self.queue.get(False)
				self.idle.clear()
			except:
				# no work to do
				# if not self.idle.is_set():
				#  print >> stdout, '%s is idle' % self.name
				self.idle.set()
				continue

			try:
				# the function may raise
				result = func(*args, **kwargs)
				if result is not None:
					self.results.put(result)
			except Exception as e:
				# so we move on and handle it in whatever way the caller wanted
				self.exception_handler(self.name, e, args, kwargs)
			finally:
				# task complete no matter what happened
				self.queue.task_done()


# class for thread pool
class Pool:

	"""Pool of threads consuming tasks from a queue"""

	def __init__(self, thread_count, batch_mode=False, exception_handler=default_handler):
		# batch mode means block when adding tasks if no threads available to process
		self.queue = Queue(thread_count if batch_mode else 0)
		self.resultQueue = Queue(0)
		self.thread_count = thread_count
		self.exception_handler = exception_handler
		self.aborts = []
		self.idles = []
		self.threads = []

	"""Tell my threads to quit"""

	def __del__(self):
		self.abort()

	"""Start the threads, or restart them if you've aborted"""

	def run(self, block=False):
		# Either wait for them to finish or return false if some aren't
		if block:
			while self.alive():
				sleep(1)
		elif self.alive():
			return False

		# Start them
		self.aborts = []
		self.idles = []
		self.threads = []
		for n in range(self.thread_count):
			abort = Event()
			idle = Event()
			self.aborts.append(abort)
			self.idles.append(idle)
			self.threads.append(Worker("thread-%d" % n, self.queue, self.resultQueue, abort, idle, self.exception_handler))
		return True

	"""Add a task to the queue"""

	def enqueue(self, func, *args, **kargs):
		self.queue.put((func, args, kargs))

	"""Wait for completion of all the tasks in the queue"""

	def join(self):
		self.queue.join()

	"""Tell each worker that its done working"""

	def abort(self, block=False):
		# tell the threads to stop after they are done with what they are currently doing
		for a in self.aborts:
			a.set()
		# wait for them to finish if requested
		while block and self.alive():
			sleep(1)

	"""Returns True if any threads are currently running"""

	def alive(self):
		return True in [t.is_alive() for t in self.threads]

	"""Returns True if all threads are waiting for work"""

	def idle(self):
		return False not in [i.is_set() for i in self.idles]

	"""Returns True if not tasks are left to be completed"""

	def done(self):
		return self.queue.empty()

	"""Get the set of results that have been processed, repeatedly call until done"""

	def results(self, wait=0):
		sleep(wait)
		results = []
		try:
			while True:
				# get a result, raises empty exception immediately if none available
				results.append(self.resultQueue.get(False))
				self.resultQueue.task_done()
		except:
			pass
		return results


# run the test
if __name__ == "__main__":

	def work(x):
		sleep(5)
		print >> stdout, "%d finished" % x
		return x

	def wait_for_results(p):
		while not p.done() or not p.idle():
			for result in p.results():
				print >> stdout, "got result %s" % str(result)
		for result in p.results():
			print >> stdout, "got result %s" % str(result)
		p = Pool(3)
		print >> stdout, "queueing work"
		for i in range(10):
			p.enqueue(work, i)
		print >> stdout, "starting work"
		p.run(True)
		print >> stdout, "waiting"
		sleep(8)
		for result in p.results():
			print >> stdout, "got result %s" % str(result)
		print >> stdout, "cancel unstarted work"
		p.abort()
		print >> stdout, "waiting and restarting"
		p.run(True)
		print >> stdout, "restarted waiting for results"
		wait_for_results(p)
		print >> stdout, "adding more work"
		for i in range(10, 17):
			p.enqueue(work, i)
		wait_for_results(p)



from time import time
import datetime
import math


def _human_time(*args, **kwargs):
	secs = float(datetime.timedelta(*args, **kwargs).total_seconds())
	units = [("day", 86400), ("hour", 3600), ("minute", 60), ("second", 1)]
	parts = []
	for unit, mul in units:
		if secs / mul >= 1 or mul == 1:
			if mul > 1:
				n = int(math.floor(secs / mul))
				secs -= n * mul
			else:
				n = secs if secs != int(secs) else int(secs)
			parts.append("%s %s%s" % (n, unit, "" if n == 1 else "s"))
	return ", ".join(parts)


class Profiler:
	def __init__(self, message):
		self.message = message

	def __enter__(self):
		self.start = time()
		return None  # could return anything, to be used like this: with Timer("Message") as value:

	def __exit__(self, type, value, traceback):
		sec = time() - self.start
		elapsed_time = _human_time(seconds=sec)
		print(f"{self.message} took {format(elapsed_time)}")
